@@ -102,10 +102,16 @@ elif [ "${machine}" = "ia64" ]; then
     verity_guid=86ed10d5-b607-45bb-8957-d350f23d0571
     signature_guid=e98b36ee-32ba-4882-9b12-0ce14655f46a
     architecture="ia64"
+elif [ "${machine}" = "s390x" ]; then
+    root_guid=5eead9a9-fe09-4a1e-a1d7-520d00531306
+    verity_guid=b325bfbe-c7be-4ab8-8357-139e652d2f6b
+    signature_guid=c80187a5-73a3-491a-901a-017c3fa953e9
+    architecture="s390x"
 elif [ "${machine}" = "ppc64le" ]; then
-    # There's no support of PPC in the discoverable partitions specification yet, so skip the rest for now
-    echo OK >/testok
-    exit 0
+    root_guid=c31c45e6-3f39-412e-80fb-4809c4980599
+    verity_guid=906bd944-4589-4aae-a4e4-dd983917446a
+    signature_guid=d4a236e7-e873-4c07-bf1d-bf6cf7f1c3c6
+    architecture="ppc64-le"
 else
     echo "Unexpected uname -m: ${machine} in testsuite-50.sh, please fix me"
     exit 1
@@ -181,12 +187,19 @@ if [ "${HAVE_OPENSSL}" -eq 1 ]; then
     sfdisk --part-label "${image}.gpt" 3 "Signature Partition"
 fi
 loop="$(losetup --show -P -f "${image}.gpt")"
+# kernel sometimes(?) does not emit add uevent for partitions, let's wait for the whole block device.
+udevadm wait --timeout 60 --settle "${loop:?}"
 dd if="${image}.raw" of="${loop}p1"
 dd if="${image}.verity" of="${loop}p2"
 if [ "${HAVE_OPENSSL}" -eq 1 ]; then
     dd if="${image}.verity-sig" of="${loop}p3"
 fi
 losetup -d "${loop}"
+# Interestingly, kernel does not emit remove uevent for the whole block device, but does for partitions.
+udevadm wait --timeout 60 --settle --removed "${loop}p1" "${loop}p2"
+if [ "${HAVE_OPENSSL}" -eq 1 ]; then
+    udevadm wait --timeout 60 --settle --removed "${loop}p3"
+fi
 
 # Derive partition UUIDs from root hash, in UUID syntax
 ROOT_UUID="$(systemd-id128 -u show "$(head -c 32 "${image}.roothash")" -u | tail -n 1 | cut -b 6-)"
@@ -279,7 +292,7 @@ Type=notify
 RemainAfterExit=yes
 MountAPIVFS=yes
 PrivateTmp=yes
-ExecStart=/bin/sh -c 'systemd-notify --ready; while ! grep -q -F MARKER /tmp/img/usr/lib/os-release; do sleep 0.1; done; mount | grep -F "/tmp/img" | grep -q -F "nosuid"'
+ExecStart=/bin/sh -c 'systemd-notify --ready; while ! grep -q -F MARKER /tmp/img/usr/lib/os-release; do sleep 0.1; done; mount | grep -F "/dev/mapper/${roothash}-verity" | grep -q -F "nosuid"'
 EOF
 systemctl start testservice-50d.service
 
@@ -316,6 +329,37 @@ EOF
 systemctl start testservice-50e.service
 systemctl is-active testservice-50e.service
 
+# ExtensionDirectories will set up an overlay
+mkdir -p "${image_dir}/app0" "${image_dir}/app1"
+systemd-run -P --property ExtensionDirectories="${image_dir}/nonexistent" --property RootImage="${image}.raw" cat /opt/script0.sh && { echo 'unexpected success'; exit 1; }
+systemd-run -P --property ExtensionDirectories="${image_dir}/app0" --property RootImage="${image}.raw" cat /opt/script0.sh && { echo 'unexpected success'; exit 1; }
+systemd-dissect --mount /usr/share/app0.raw "${image_dir}/app0"
+systemd-dissect --mount /usr/share/app1.raw "${image_dir}/app1"
+systemd-run -P --property ExtensionDirectories="${image_dir}/app0" --property RootImage="${image}.raw" cat /opt/script0.sh | grep -q -F "extension-release.app0"
+systemd-run -P --property ExtensionDirectories="${image_dir}/app0" --property RootImage="${image}.raw" cat /usr/lib/systemd/system/some_file | grep -q -F "MARKER=1"
+systemd-run -P --property ExtensionDirectories="${image_dir}/app0 ${image_dir}/app1" --property RootImage="${image}.raw" cat /opt/script0.sh | grep -q -F "extension-release.app0"
+systemd-run -P --property ExtensionDirectories="${image_dir}/app0 ${image_dir}/app1" --property RootImage="${image}.raw" cat /usr/lib/systemd/system/some_file | grep -q -F "MARKER=1"
+systemd-run -P --property ExtensionDirectories="${image_dir}/app0 ${image_dir}/app1" --property RootImage="${image}.raw" cat /opt/script1.sh | grep -q -F "extension-release.app2"
+systemd-run -P --property ExtensionDirectories="${image_dir}/app0 ${image_dir}/app1" --property RootImage="${image}.raw" cat /usr/lib/systemd/system/other_file | grep -q -F "MARKER=1"
+cat >/run/systemd/system/testservice-50f.service <<EOF
+[Service]
+MountAPIVFS=yes
+TemporaryFileSystem=/run /var/lib
+StateDirectory=app0
+RootImage=${image}.raw
+ExtensionDirectories=${image_dir}/app0 ${image_dir}/app1
+# Relevant only for sanitizer runs
+UnsetEnvironment=LD_PRELOAD
+ExecStart=/bin/bash -c '/opt/script0.sh | grep ID'
+ExecStart=/bin/bash -c '/opt/script1.sh | grep ID'
+Type=oneshot
+RemainAfterExit=yes
+EOF
+systemctl start testservice-50f.service
+systemctl is-active testservice-50f.service
+umount "${image_dir}/app0"
+umount "${image_dir}/app1"
+
 echo OK >/testok
 
 exit 0
